/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.utl;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.mockito.cglib.proxy.Enhancer;

import xgi.ut.dsl.annot.ConversionMethod;
/**
 * This class provides a convenient workaround to avoid the fact that 
 * scala arrays are not automatically expanded to vararg when it comes
 * to java.lang.reflect.Method invocation
 */
public class ReflectionUtil {
	/** This field holds a cache of all the fields of all the previously inspected classes */
	private static final Map<Class<?>, Map<String, Field>> FIELD_CACHE = new HashMap<Class<?>, Map<String,Field>>();
	/** This field holds a cache of all the methods of all the previously inspected classes */
	private static final Map<Class<?>, Map<String, List<Method>>> METHOD_CACHE = new HashMap<Class<?>, Map<String,List<Method>>>();
	/**	Pertains the assignable from relationship */
	private static final Map<Class, List<Class>> BOXING_MATRIX = new HashMap<Class, List<Class>>();
	static{
		BOXING_MATRIX.put(double.class, Arrays.asList(new Class[]{char.class, byte.class, short.class, int.class, long.class, float.class, double.class}));
		BOXING_MATRIX.put(float.class, Arrays.asList( new Class[]{char.class, byte.class, short.class, int.class, long.class, float.class}));
		BOXING_MATRIX.put(long.class, Arrays.asList(  new Class[]{char.class, byte.class, short.class, int.class, long.class}));
		BOXING_MATRIX.put(int.class, Arrays.asList(   new Class[]{char.class, byte.class, short.class, int.class}));
		BOXING_MATRIX.put(short.class, Arrays.asList( new Class[]{char.class, byte.class, short.class}));
		BOXING_MATRIX.put(byte.class, Arrays.asList(  new Class[]{char.class, byte.class}));
		BOXING_MATRIX.put(boolean.class, Arrays.asList(new Class[]{boolean.class}));
	}
	
	/** 
	 * invokes the method identified by name on object target with the given parameters 
	 * and assigns it a mock result.
	 * @param target the object on which to invoke the method
	 * @param name the name of the method to invoke
	 * @param params the method parameters
	 * @param result the mock result to assign to the fn call
	 * @return the result of the fn call
	 * @throws NoSuchMethodException in case the method does not exists in the class
	 * @throws IllegalAccessException in case the method is not accessible
	 * @throws InvocationTargetException in case the method cannot be called
	 */
	public static final Object invoke(Object target, String name,
			List<?> params, Object result) throws NoSuchMethodException,
			IllegalAccessException, InvocationTargetException {
		Object invocation = null;
		Class<?> clazz = target.getClass();
		// was original class a stub ? 
		boolean wasOriginalClassAStub = Enhancer.isEnhanced(clazz);
		
		// retrieve the method
		Method m = getMatchingMethod(clazz, name, params);

		// Invoke the method		
		if (!wasOriginalClassAStub) {
			try {
				invocation = m.invoke(target, boxArgs(params.toArray(), m.getParameterTypes()));
			} catch (Throwable e) {
				// swallow
			}
		} else {
//			Object fnRes = result;
//			if (fnRes == null) {
//				fnRes = mock(m.getReturnType());
//			}
			try {
				// record the call in mockito
				when(m.invoke(target, boxArgs(params.toArray(), m.getParameterTypes()))).thenReturn(result);
			} catch (Exception e) {
				e.printStackTrace();
			}
			invocation = result;
		}
		return invocation;
	}
	
	/** 
	 * invokes the method identified by name on object target with the given parameters 
	 * and assigns it a mock result.
	 * @param target the object on which to invoke the method
	 * @param name the name of the method to invoke
	 * @param params the method parameters
	 * @param result the mock result to assign to the fn call
	 * @return the result of the fn call
	 * @throws NoSuchMethodException in case the method does not exists in the class
	 * @throws IllegalAccessException in case the method is not accessible
	 * @throws InvocationTargetException in case the method cannot be called
	 */
	public static final void invokeException(Object target, String name,
			List<?> params, Throwable result) throws NoSuchMethodException,
			IllegalAccessException, InvocationTargetException {
		Class<?> clazz = target.getClass();
		// was original class a stub ? 
		boolean wasOriginalClassAStub = Enhancer.isEnhanced(clazz);
		
		// retrieve the method
		Method m = getMatchingMethod(clazz, name, params);

		// Invoke the method		
		if (!wasOriginalClassAStub) {
			throw new IllegalStateException("You cannot configure an exception on an object that is not a mock");
		} else {
			try {
				// record the call in mockito
				when(m.invoke(target, boxArgs(params.toArray(), m.getParameterTypes()))).thenThrow(result);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * This method provides a convenient way to invoke the conversion method of a conversion class
	 * @param text the text that needs to be converted (parsed)
	 * @param conversionClass the class that contains one conversion method (public static annotated with @ConversionMethod)
	 * @return the object that is represented by text
	 * 
	 * @throws IllegalArgumentException in case there is a mismatch btw the given arguments and that of the conversion method
	 */
	public static final Object invokeConversion(String text, Class<?> conversionClass) throws IllegalArgumentException {
		Method[] methods = conversionClass.getMethods();
		for(Method m : methods){
			ConversionMethod annotation = m.getAnnotation(ConversionMethod.class);
			if(annotation!=null){
				if(!Modifier.isStatic(m.getModifiers()))throw new RuntimeException("The conversion method must be static");
				if(!Modifier.isPublic(m.getModifiers()))throw new RuntimeException("The conversion method must be public");
				try {
					return m.invoke(null, text);
				} catch (IllegalAccessException e) {
					//swallow : it cannot happen
				} catch (InvocationTargetException e) {
					//swallow : it cannot happen
				}
			}
		}
		throw new RuntimeException("The parser class "+conversionClass.getCanonicalName()+" contains no public static method annotated with @ConversionMethod");
	}
	
	/**
	 * Retrieves all the public + private (declared + inherited) fields of a given class
	 * 
	 * Note: as a side effect, this method will populate the CACHE field of this class.
	 * 
	 * @param clazz the class to inspect
	 * @return a map with the key being the field name and the value the field itself.
	 */
	public static Map<String, Field> getAllFields(Class<?> clazz){
		Map<String, Field> results = FIELD_CACHE.get(clazz);
		if(results!=null){
			return results;
		}
		if(clazz.equals(Object.class)){
			return new HashMap<String, Field>();
		}
		Class<?> parent = clazz.getSuperclass();
		results = new HashMap<String, Field>(getAllFields(parent));
		for(Field f : clazz.getDeclaredFields()){
			results.put(f.getName(), f);
		}
		FIELD_CACHE.put(clazz, results);
		return results;
	}
	
	/**
	 * Retrieves all the public + private (declared + inherited) methods of a given class
	 * 
	 * Note : as a side effect, this method will populate the method CACHE of this class.
	 * Note2: mock classes are not taken into account
	 * 
	 * @param clazz the class to inspect
	 * @return a map with the key being the method name and the value the list of all methods named the same way.
	 */
	public static Map<String, List<Method>> getAllMethods(Class<?> clazz){
		Map<String, List<Method>> results = METHOD_CACHE.get(clazz);
		if(results!=null){
			return results;
		}
		
		Class<?> parent = clazz.getSuperclass();
		results = new HashMap<String, List<Method>>();
		if(parent!=null){
			results.putAll(getAllMethods(parent));
		}
		for(Method m : clazz.getDeclaredMethods()){
			List<Method> methods = results.get(m.getName());
			if(methods == null){
				methods = new ArrayList<Method>();
				results.put(m.getName(), methods);
			}
			methods.add(m);
		}
		METHOD_CACHE.put(clazz, results);
		return results;
	}
	
	/**
	 * Retrieves a method that has a signature compatible with the given name and parameters
	 * @param clazz the class to inspect
	 * @param name the method name
	 * @param args the method arguments
	 * @return the method
	 */
	public static Method getMatchingMethod(Class<?> clazz, String name, List<?> args){
		List<Method> possibilities = getAllMethods(clazz).get(name);
		if(possibilities == null || possibilities.isEmpty()){
			throw new RuntimeException(name+Arrays.asList(args.toArray())+" not found in "+clazz);
		}
		
		for(Method m : possibilities){
			Class<?>[] paramTypes = m.getParameterTypes();
			if(args.size()!=paramTypes.length) continue;
			
			boolean isMatch = true;
			for(int i=0; i<paramTypes.length; i++){
				Object arg = args.get(i);
				if(arg==null) continue;
				isMatch &= isAssignableFrom(paramTypes[i],arg.getClass());
			}
			if(isMatch){
				return m;
			}
		}
		throw new RuntimeException(name+Arrays.asList(args.toArray())+" not found in "+clazz);
	}
	
	/**
	 * Determines if a type a can be assigned/auto-boxed to a type b
	 * @param a a type
	 * @param b a type
	 * @return true if a can be assigned/auto-boxed to type b, false otherwise
	 */
	private static boolean isAssignableFrom(Class<?> a, Class<?> b){
		boolean assignable = false;
		if(a.isPrimitive() && b.isPrimitive()){
			assignable = BOXING_MATRIX.get(a).contains(b);
		} else if((a.isPrimitive() ^ b.isPrimitive()) || (Number.class.isAssignableFrom(a) && Number.class.isAssignableFrom(b))){
			assignable = isAssignableFrom(unwrap(a), unwrap(b));
		} else {
			assignable = a.isAssignableFrom(b);
		}
		return assignable;
	}
	/**
	 * Returns the native type corresponding to type a if a is a primitive type
	 * @param a the type to unwrap
	 * @return true if a is the wrapper to a primitive type
	 */
	private static Class<?> unwrap(Class<?> a){
		if(Character.class.equals(a)){
			return char.class;
		}
		if(Byte.class.equals(a)){
			return byte.class;
		}
		if(Short.class.equals(a)){
			return short.class;
		}
		if(Integer.class.equals(a)){
			return int.class;
		}
		if(Long.class.equals(a)){
			return long.class;
		}
		if(Float.class.equals(a)){
			return float.class;
		}
		if(Double.class.equals(a)){
			return double.class;
		}
		if(Boolean.class.equals(a)){
			return boolean.class;
		}
		return a;
	}
	
	/**
	 * Returns the value o boxed in the type targetType if an autoboxing is possible
	 * @param o the object to autobox
	 * @param targetType the type to which o should be boxed
	 * @return the value o boxed in the type targetType if an autoboxing is possible
	 */
	public static Object autoBox(Object o, Class<?> targetType){
		// not boxable
		if(!(Number.class.isAssignableFrom(targetType) || targetType.isPrimitive())){
			return o;
		}
		if(!(o instanceof Number)){
			return o;
		}
		Number n = (Number)o;
		if(byte.class.equals(unwrap(targetType))){
			return n.byteValue();
		}
		if(short.class.equals(unwrap(targetType))){
			return n.shortValue();
		}
		if(int.class.equals(unwrap(targetType))){
			return n.intValue();
		}
		if(long.class.equals(unwrap(targetType))){
			return n.longValue();
		}
		if(float.class.equals(unwrap(targetType))){
			return n.floatValue();
		}
		if(double.class.equals(unwrap(targetType))){
			return n.doubleValue();
		}
		// how did i end up here ??
		return n;
	}
	
	/**
	 * Applies an autoboxing to all the arguments of a function call
	 * @param args the array of parameters
	 * @param paramTypes the type of the fn params
	 * @return the parameters that all have been autoboxed
	 */
	private static Object[] boxArgs(Object[]args, Class<?>[] paramTypes){
		Object[] ret = new Object[args.length];
		for(int i = 0; i<args.length; i++){
			ret[i] = autoBox(args[i], paramTypes[i]);
		}
		return ret;
	}
}
